from libcpp.deque cimport deque

ctypedef deque[ActhonMessageInterface] message_queue_t

cdef extern from "<semaphore.h>" nogil:
  ctypedef int sem_t
  int sem_init(sem_t *sem, int pshared, unsigned int value)
  int sem_wait(sem_t *sem)
  int sem_post(sem_t *sem)
  int sem_destroy(sem_t* sem)


cdef cypclass BasicQueue(ActhonQueueInterface) checklock:
  message_queue_t* _queue

  __init__(self):
    self._queue = new message_queue_t()

  __dealloc__(self):
    del self._queue

  bint is_empty(self) const:
    return self._queue.empty()

  void push(self, ActhonMessageInterface message):
    self._queue.push_back(message)
    if message._sync_method is not NULL:
      message._sync_method.insertActivity(message)
    Cy_INCREF(message)

  bint activate(self):
    cdef bint one_message_processed
    if self._queue.empty():
      return False
    # Note here that according to Cython refcount conventions,
    # the front() method should have returned a new ref.
    # This is obviously not the case, so if we do nothing
    # we will, at the end of this function, loose a ref on the pointed object
    # (as we will decref the thing pointed by next_message).
    next_message = self._queue.front()
    self._queue.pop_front()
    one_message_processed = next_message.activate()
    if one_message_processed:
      if next_message._sync_method is not NULL:
        with wlocked next_message._sync_method:
          next_message._sync_method.removeActivity(next_message)
    else:
      self._queue.push_back(next_message)
      # Don't forget to incref to avoid premature deallocation
      Cy_INCREF(next_message)
    return one_message_processed

cdef cypclass NoneResult(ActhonResultInterface) checklock:
  void pushVoidStarResult(self, void* result):
    pass
  void pushIntResult(self, int result):
    pass
  void* getVoidStarResult(self) const:
    return NULL
  int getIntResult(self) const:
    return 0
  @staticmethod
  ActhonResultInterface construct():
    return NoneResult()

cdef cypclass WaitResult(ActhonResultInterface) checklock:
  union result_t:
    int int_val
    void* ptr
  result_t result
  sem_t semaphore

  __init__(self):
    self.result.ptr = NULL
    sem_init(&self.semaphore, 0, 0)

  __dealloc__(self):
    sem_destroy(&self.semaphore)

  @staticmethod
  ActhonResultInterface construct():
    return WaitResult()

  void pushVoidStarResult(self, void* result):
    self.result.ptr = result
    sem_post(&self.semaphore)

  void pushIntResult(self, int result):
    self.result.int_val = result
    sem_post(&self.semaphore)

  result_t _getRawResult(self) const:
    # We must ensure a result exists, but we can let others access it immediately
    # The cast here is a way of const-casting (we're modifying the semaphore in a const method)
    sem_wait(<sem_t*> &self.semaphore)
    sem_post(<sem_t*> &self.semaphore)
    return self.result

  void* getVoidStarResult(self) const:
    res = self._getRawResult()
    return res.ptr

  int getIntResult(self) const:
    res = self._getRawResult()
    return res.int_val

cdef cypclass SimpleResult(ActhonResultInterface) nolock:
  """
  This is a wrapper around BasicResult that is handling locking the underlying object for each function call.
  Consequently this is a way to emulate autolock mode without the deadlock issues this mode has with acthon.
  If you have many calls to this object, things can be slower (you're locking / unlocking for each call)
  Moreover, the cython-generated part uses this to push the result of an asynchronous execution
  of a function. This cython-generated will lock and unlock: it has no way to guess
  the result interface you're providing doesn't need locking. So, for the result pushing,
  you'll have write locking and unlocking on the BasicResult object AND on the whole SimpleResult object.

  So this is simpler to use, but definitely slower than BasicResult.
  """
  WaitResult member
  __init__(self):
    self.member = WaitResult()

  @staticmethod
  ActhonResultInterface construct():
    return SimpleResult()

  void pushVoidStarResult(self, void* result):
    with wlocked self.member:
      self.member.pushVoidStarResult(result)

  void pushIntResult(self, int result):
    with wlocked self.member:
      self.member.pushIntResult(result)

  void* getVoidStarResult(self) const:
    cdef void* res
    with rlocked self.member:
      res = self.member.getVoidStarResult()
    return res

  int getIntResult(self) const:
    cdef int res
    with rlocked self.member:
      res = self.member.getIntResult()
    return res

cdef cypclass ActivityCounterSync(ActhonSyncInterface) checklock:
  int count
  ActivityCounterSync previous_sync

  __init__(self, ActivityCounterSync prev = <ActivityCounterSync> NULL):
    self.count = 0
    self.previous_sync = prev

  void insertActivity(self, ActhonMessageInterface msg):
    self.count += 1

  void removeActivity(self, ActhonMessageInterface msg):
    self.count -= 1

  bint isCompleted(self) const:
    return self.count == 0

  bint isActivable(self) const:
    cdef bint res = True
    if self.previous_sync is not NULL:
      with rlocked self.previous_sync:
        res = self.previous_sync.isCompleted()
    return res


cdef cypclass BasicActivableClass autolock activable:
    __init__(self):
        self._active_queue_class = BasicQueue()
        self._active_result_class = SimpleResult.construct
